package pkg

import (
	"context"
	"fmt"
	"github.com/julienschmidt/httprouter"
	"html/template"
	"math"
	"net/http"
	"strconv"
)

type ProductEntry struct {
	Slug  string
	Name  string
	Price string
	Tags  []string
}

type productsIndexPage struct {
	Products []*ProductEntry
}

func IndexProducts(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	conn := getConn(r)
	company := mustGetCompany(r)
	tag := r.URL.Query().Get("tag")
	page := &productsIndexPage{
		Products: mustCollectProductEntries(r.Context(), conn, company, tag),
	}
	mustRenderMainTemplate(w, r, "products/index.gohtml", page)
}

func GetProductForm(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
	locale := getLocale(r)
	conn := getConn(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	slug := params[0].Value
	if slug == "new" {
		w.WriteHeader(http.StatusOK)
		mustRenderNewProductForm(w, r, form)
		return
	}
	if !form.MustFillFromDatabase(r.Context(), conn, slug) {
		http.NotFound(w, r)
		return
	}
	w.WriteHeader(http.StatusOK)
	mustRenderEditProductForm(w, r, slug, form)
}

func mustRenderNewProductForm(w http.ResponseWriter, r *http.Request, form *productForm) {
	mustRenderModalTemplate(w, r, "products/new.gohtml", form)
}

type editProductPage struct {
	Slug        string
	ProductName string
	Form        *productForm
}

func mustRenderEditProductForm(w http.ResponseWriter, r *http.Request, slug string, form *productForm) {
	page := &editProductPage{
		Slug:        slug,
		ProductName: form.Name.Val,
		Form:        form,
	}
	mustRenderModalTemplate(w, r, "products/edit.gohtml", page)
}

func HandleAddProduct(w http.ResponseWriter, r *http.Request, _ httprouter.Params) {
	conn := getConn(r)
	locale := getLocale(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	if !form.Validate() {
		if !IsHTMxRequest(r) {
			w.WriteHeader(http.StatusUnprocessableEntity)
		}
		mustRenderNewProductForm(w, r, form)
		return
	}
	taxes := mustSliceAtoi(form.Tax.Selected)
	conn.MustExec(r.Context(), "select add_product($1, $2, $3, $4, $5, $6)", company.Id, form.Name, form.Description, form.Price, taxes, form.Tags)
	if IsHTMxRequest(r) {
		w.Header().Set(HxTrigger, "closeModal")
		w.Header().Set(HxRefresh, "true")
		w.WriteHeader(http.StatusNoContent)
	} else {
		http.Redirect(w, r, companyURI(company, "/products"), http.StatusSeeOther)
	}
}

func sliceAtoi(s []string) ([]int, error) {
	var i []int
	for _, vs := range s {
		vi, err := strconv.Atoi(vs)
		if err != nil {
			return i, err
		}
		i = append(i, vi)
	}
	return i, nil
}

func mustSliceAtoi(s []string) []int {
	i, err := sliceAtoi(s)
	if err != nil {
		panic(err)
	}
	return i
}

func HandleUpdateProduct(w http.ResponseWriter, r *http.Request, params httprouter.Params) {
	conn := getConn(r)
	locale := getLocale(r)
	company := mustGetCompany(r)
	form := newProductForm(r.Context(), conn, locale, company)
	if err := form.Parse(r); err != nil {
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}
	if err := verifyCsrfTokenValid(r); err != nil {
		http.Error(w, err.Error(), http.StatusForbidden)
		return
	}
	slug := params[0].Value
	if !form.Validate() {
		w.WriteHeader(http.StatusUnprocessableEntity)
		mustRenderEditProductForm(w, r, slug, form)
		return
	}
	taxes := mustSliceAtoi(form.Tax.Selected)
	if ok := conn.MustGetBool(r.Context(), "select edit_product($1, $2, $3, $4, $5, $6)", slug, form.Name, form.Description, form.Price, taxes, form.Tags); !ok {
		http.NotFound(w, r)
	}
	if IsHTMxRequest(r) {
		w.Header().Set(HxTrigger, "closeModal")
		w.Header().Set(HxRefresh, "true")
		w.WriteHeader(http.StatusNoContent)
	} else {
		http.Redirect(w, r, companyURI(company, "/products/"+slug), http.StatusSeeOther)
	}
}

func mustCollectProductEntries(ctx context.Context, conn *Conn, company *Company, tag string) []*ProductEntry {
	rows, err := conn.Query(ctx, `
		select product.slug
		     , product.name
		     , to_price(price, decimal_digits)
		     , tags
		from product
		join company using (company_id)
		join currency using (currency_code)
		where product.company_id = $1 and (($2 = '') or (tags @> array[$2]::tag_name[]))
		order by name
	`, company.Id, tag)
	if err != nil {
		panic(err)
	}
	defer rows.Close()

	var entries []*ProductEntry
	for rows.Next() {
		entry := &ProductEntry{}
		err = rows.Scan(&entry.Slug, &entry.Name, &entry.Price, &entry.Tags)
		if err != nil {
			panic(err)
		}
		entries = append(entries, entry)
	}
	if rows.Err() != nil {
		panic(rows.Err())
	}

	return entries
}

type productForm struct {
	locale      *Locale
	company     *Company
	Name        *InputField
	Description *InputField
	Price       *InputField
	Tax         *SelectField
	Tags        *TagsField
}

func newProductForm(ctx context.Context, conn *Conn, locale *Locale, company *Company) *productForm {
	return &productForm{
		locale:  locale,
		company: company,
		Name: &InputField{
			Name:     "name",
			Label:    pgettext("input", "Name", locale),
			Type:     "text",
			Required: true,
		},
		Description: &InputField{
			Name:  "description",
			Label: pgettext("input", "Description", locale),
			Type:  "textarea",
		},
		Price: &InputField{
			Name:     "price",
			Label:    pgettext("input", "Price", locale),
			Type:     "number",
			Required: true,
			Attributes: []template.HTMLAttr{
				`min="0"`,
				template.HTMLAttr(fmt.Sprintf(`step="%v"`, company.MinCents())),
			},
		},
		Tax: &SelectField{
			Name:     "tax",
			Label:    pgettext("input", "Taxes", locale),
			Multiple: true,
			Options:  mustGetTaxOptions(ctx, conn, company),
		},
		Tags: &TagsField{
			Name:  "tags",
			Label: pgettext("input", "Tags", locale),
		},
	}
}

func (form *productForm) Parse(r *http.Request) error {
	if err := r.ParseForm(); err != nil {
		return err
	}
	form.Name.FillValue(r)
	form.Description.FillValue(r)
	form.Price.FillValue(r)
	form.Tax.FillValue(r)
	form.Tags.FillValue(r)
	return nil
}

func (form *productForm) Validate() bool {
	validator := newFormValidator()
	validator.CheckRequiredInput(form.Name, gettext("Name can not be empty.", form.locale))
	if validator.CheckRequiredInput(form.Price, gettext("Price can not be empty.", form.locale)) {
		validator.CheckValidDecimal(form.Price, form.company.MinCents(), math.MaxFloat64, gettext("Price must be a number greater than zero.", form.locale))
	}
	validator.CheckValidSelectOption(form.Tax, gettext("Selected tax is not valid.", form.locale))
	validator.CheckAtMostOneOfEachGroup(form.Tax, gettext("You can only select a tax of each class.", form.locale))
	return validator.AllOK()
}

func (form *productForm) MustFillFromDatabase(ctx context.Context, conn *Conn, slug string) bool {
	return !notFoundErrorOrPanic(conn.QueryRow(ctx, `
		select product.name
		     , product.description
		     , to_price(price, decimal_digits)
		     , array_agg(tax_id)
		     , array_to_string(tags, ',')
		from product
		left join product_tax using (product_id)
		join company using (company_id)
		join currency using (currency_code)
		where product.slug = $1
		group by product_id
		       , product.name
		       , product.description
		       , price
		       , tags
		       , decimal_digits
	`, slug).Scan(
		form.Name,
		form.Description,
		form.Price,
		form.Tax,
		form.Tags))
}
